# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/compat.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/compat.ipynb 1
try:
    from pyspark.sql import DataFrame as SparkDataFrame
except ImportError:

    class SparkDataFrame: ...
